/**
   \class CPersistFactory

   \brief Implements persistence for IPersistObject(s)
*/
// --------------------------------------------------------
//  Author : Sylvain Lefebvre
//  History: Created ... 28/12/2001
// --------------------------------------------------------
#ifndef __PERSISTENT_FACTORY__
#define __PERSISTENT_FACTORY__
// --------------------------------------------------------
#define PERSIST_FACTORY_MAJOR 0
#define PERSIST_FACTORY_MINOR 80
// --------------------------------------------------------
#define PERSIST_FILE_NAME_MAX_SIZE 512
// --------------------------------------------------------
#define PERSIST (*(CPersistFactory::getInstance()))
// --------------------------------------------------------
#ifdef PERSIST_XML
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/parserInternals.h>
#endif
// --------------------------------------------------------
#include <iostream>
#include <list>
#include <vector>
#include <map>
#include <string>
// --------------------------------------------------------
#include "CException.h"
#include "IPersistType.h"
#include "persist_string.h"
// --------------------------------------------------------
class IPersistObject;
class IPersistType;
// --------------------------------------------------------
class CPersistFactory
{
private:
  
  class CStrcmp
  {
  public:
    bool operator()(const char* s1, const char* s2) const
      {
	return strcmp(s1, s2) < 0;
      }
  };

public:
  class CPtrStock
  {
  private:
  public:
    virtual void set(IPersistObject *)  {}
  };
  
  template <class T> class CPtrHack : public CPtrStock
  {
  private:
    T             **m_Ptr;
  public:
    CPtrHack()                           {m_Ptr=NULL;}
    CPtrHack(T **p) : CPtrStock()        {m_Ptr=p;}
    virtual void set(IPersistObject *p) 
      {
	if (p != NULL)
	  *m_Ptr=dynamic_cast<T *>(p);
	else
	  *m_Ptr=NULL;
      }
  };

private:

  vector<IPersistObject *>                         m_Registred;
  vector<list<CPtrStock *> >                       m_Needed;  
  list<IPersistObject *>                           m_Loaded;
  static map<const char *,IPersistType *,CStrcmp> *s_Types;
  static persist_string                            s_AppVersion;
  vector<IPersistObject *>                         m_RegistredData;
  bool                                             m_bLoadInProgress;
  int                                              m_iNbObjectsLoaded;
  bool                                             m_bBeforeFirstLoad;
  bool                                             m_bMerge;
  int                                              m_iIdsOffset;
  char                                             m_szCurrentFileName[PERSIST_FILE_NAME_MAX_SIZE];

  CPersistFactory();
  ~CPersistFactory();
  
  //! register an object as persistent
  void               registerObject(IPersistObject *);
  //! register an object's data as persistent
  void               registerData(IPersistObject *);
  //! unregister an object's data
  void               unregisterData(IPersistObject *);  
  //! unregister an object
  void				       unregisterObject(IPersistObject *);
  //! compact ids to avoid NULL ptrs in m_Registred
  void               compactIds();
  //! register a new type
  static void        registerType(IPersistType *);

  friend class IPersistObject;
  friend class IPersistType;

protected:

  //! save an object on stream
  void               rawSaveObject(ostream&,IPersistObject *);
  //! build an object from stream
  IPersistObject    *rawBuildObject(istream& );
  //! save persistent objects
  void               rawSaveObjects(ostream&);
  //! save persistent datas
  void               rawSaveDatas(ostream&);
  //! load persistent objects
  void               rawLoadObjects(istream&);
  //! load persistent datas
  void               rawLoadDatas(istream&);

#ifdef PERSIST_XML
  //! save an object on stream in xml format
  void               xmlSaveObject(ostream&,IPersistObject *);
  //! save persistent objects in xml format
  void               xmlSaveObjects(ostream&);
  //! save persistent datas in xml format
  void               xmlSaveDatas(ostream&);
  //! parse and load xml persist datas
  void               xmlLoadLoop(xmlNodePtr node,IPersistObject *instance=NULL);
  void               xmlLoadLoopElement(xmlNodePtr node,vector<string>& strs);
  void               xmlLoadLoopPTRElement(xmlNodePtr node,vector<string>& ids,vector<bool>& idstype);
#endif

  //! in case of load failure, init datas with defaults
  void               defaultDatas();
  //! retrieve child pointer from id
  void               childFromId(int id,CPtrStock *);

public:
  
  //! iterator for objects retrievial
  typedef list<IPersistObject *>::iterator iterator;

  //! retrieve the unique instance of the factory
  static CPersistFactory *getInstance()
    {
      static CPersistFactory *factory=new CPersistFactory();
      
      return (factory);
    }

  //! unregister an object
  void         unregister(IPersistObject *);

  //! save persistent data to file
  void         rawSave(const char *);
  //! load persistent data from file
  void         rawLoad(const char *);
  //! called by an object when he needs pointer from child id
  void         rawLoadChild(istream&,CPtrStock *);
  //! called by an object when he need to save a child persistent object
  void         rawSaveChild(ostream&,IPersistObject *);
  //! save an instance
  bool rawSaveInstance(IPersistObject *o,const char *n);
  //! load an instance
  IPersistObject *rawLoadInstance(const char *n);

#ifdef PERSIST_XML
  //! save persistent data to file
  void         xmlSave(const char *);
  //! load persistent data from file
  void         xmlLoad(const char *);
  //! Load an instance
  IPersistObject *xmlLoadInstance(const char *n);
  //! called by an object when he need to save a child persistent object
  void         xmlSaveChild(ostream&,IPersistObject *);
  //! save one instance of a persistent object
  bool         xmlSaveInstance(IPersistObject *o,const char *n);
#endif

  void         afterLoadPoint() {m_bBeforeFirstLoad=false;}
  const char  *getCurrentFileName() const {return (m_szCurrentFileName);} 
  void         setMerge(bool b) {m_bMerge=b;}

  //! return an iterator on objects
  iterator begin() {return (m_Loaded.begin());}
  iterator end()   {return (m_Loaded.end());}

  //! set application version string (use PERSIST_APPLICATION_VERSION macro)
  void setApplicationVersionString(const string&);


};
// --------------------------------------------------------
#endif
// --------------------------------------------------------
